﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis.SolutionCaching.PermutationEncoding;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("PermutationConvexHullModifier", "Calculates the convex hull from a solution cache. ")]
  [StorableClass]
  public class PermutationConvexHullModifier : ParameterizedNamedItem, IRunCollectionModifier {
    private const string SolutionCacheResultName = "SolutionCache";

    [StorableConstructor]
    protected PermutationConvexHullModifier(bool deserializing) : base(deserializing) { }
    protected PermutationConvexHullModifier(PermutationConvexHullModifier original, Cloner cloner) : base(original, cloner) { }
    public PermutationConvexHullModifier() { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PermutationConvexHullModifier(this, cloner);
    }

    public void Modify(List<IRun> runs) {
      foreach (var run in runs) {
        int i = 0;
        if (!run.Results.ContainsKey(SolutionCacheResultName)) continue;
        var solutionCache = run.Results.Single(x => x.Key == "SolutionCache").Value as PermutationSolutionCache;
        if (solutionCache == null) continue;

        DataTable volDataTable = new DataTable("Convex hull volume over generations");
        DataTable nrOfPointsTable = new DataTable("Nr. of points of convex hull");
        DoubleValue overallVolumn = new DoubleValue();

        DataRow dtVolumeRow = new DataRow("Volume");
        DataRow dtNrPointsRow = new DataRow("Nr. of points");
        volDataTable.Rows.Add(dtVolumeRow);
        nrOfPointsTable.Rows.Add(dtNrPointsRow);

        var sols = solutionCache.GetSolutionsFromGeneration(i);
        int dim;
        double[][] dm;
        double[][] popPoints;
        while (sols.Count != 0) {
          dm = CalculateDistanceMatrixFromPermutations(sols);

          dim = sols.First().Length;
          if (dim > dm.Length)
            dim = dm.Length - 1;

          popPoints = DistanceMatrixToPoints.MetricMDS(dm, dim);

          var convexHull = LPHull.Calculate(popPoints);
          var volume = ConvexHullMeasures.CalculateVolume(convexHull);
          dtVolumeRow.Values.Add(volume);
          dtNrPointsRow.Values.Add(convexHull.Count);

          sols = solutionCache.GetSolutionsFromGeneration(++i);
        }

        //calculate convex hull for all individuals
        /* dm = CalculateDistanceMatrixFromPermutations(solutionCache.Keys().ToList());
         dim = solutionCache.Keys().First().Length;
         if (dim > dm.Length)
           dim = dm.Length - 1;
         popPoints = DistanceMatrixToPoints.MetricMDS(dm, dim);

         List<double[]> allPoints = LPHull.Calculate(popPoints);
         overallVolumn.Value = ConvexHullMeasures.CalculateVolume(allPoints);*/

        //run.Results.Add("Overall volumn: ", overallVolumn);
        run.Results.Add("Convex hull volumn", volDataTable);
        run.Results.Add("Convex nr. of points", nrOfPointsTable);
      }
    }

    private double[][] CalculateDistanceMatrixFromPermutations(List<Permutation> points) {
      double[][] tmpDm = new double[points.Count][];
      AllocArray(tmpDm, points.Count);

      for (int i = 0; i < points.Count; i++) {
        for (int j = 0; j < points.Count; j++) {
          double diversity = TSPSimilarityCalculator.CalculateSimilarity(points[i], points[j]);
          tmpDm[i][j] = diversity;
        }
      }

      return DistanceMatrixToPoints.TransformToDistances(tmpDm);
    }

    private void AllocArray(double[][] arr, int size) {
      for (int i = 0; i < arr.Length; i++) {
        arr[i] = new double[size];
      }
    }
  }
}